# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2022, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
Access to of PCI/PCIe device hierarchy
- enumerating PCI/PCIe devices
- read/write access to PCI configuration headers/registers
- enumerating PCI expansion (option) ROMs
- identifying PCI/PCIe devices MMIO and I/O ranges (BARs)

usage:
    >>> self.cs.pci.read_byte( 0, 0, 0, 0x88 )
    >>> self.cs.pci.write_byte( 0, 0, 0, 0x88, 0x1A )
    >>> self.cs.pci.enumerate_devices()
    >>> self.cs.pci.enumerate_xroms()
    >>> self.cs.pci.find_XROM( 2, 0, 0, True, True, 0xFED00000 )
    >>> self.cs.pci.get_device_bars( 2, 0, 0 )
    >>> self.cs.pci.get_DIDVID( 2, 0, 0 )
    >>> self.cs.pci.is_enabled( 2, 0, 0 )
"""

import struct
from collections import namedtuple
import itertools
from typing import List, Tuple, Optional
from chipsec.library.logger import logger, pretty_print_hex_buffer
from chipsec.library.file import write_file
from chipsec.hal.pcidb import VENDORS, DEVICES
from chipsec.library.exceptions import OsHelperError
from chipsec.library.defines import is_all_ones, MASK_16b, MASK_32b, MASK_64b, BOUNDARY_4KB

#
# PCI configuration header registers
#

# Common (type 0/1) registers
PCI_HDR_VID_OFF = 0x0
PCI_HDR_DID_OFF = 0x2
PCI_HDR_CMD_OFF = 0x4
PCI_HDR_STS_OFF = 0x6
PCI_HDR_RID_OFF = 0x8
PCI_HDR_CLSCODE_OFF = 0x9
PCI_HDR_PI_OFF = 0x9
PCI_HDR_SUB_CLS_OFF = 0xA
PCI_HDR_CLS_OFF = 0xB
PCI_HDR_CLSIZE_OFF = 0xC
PCI_HDR_MLT_OFF = 0xD
PCI_HDR_TYPE_OFF = 0xE
PCI_HDR_BIST_OFF = 0xF
PCI_HDR_CAP_OFF = 0x34
PCI_HDR_INTRLN_OFF = 0x3C
PCI_HDR_INTRPIN_OFF = 0x3D
PCI_HDR_BAR0_LO_OFF = 0x10
PCI_HDR_BAR0_HI_OFF = 0x14

# PCIe BAR register fields
PCI_HDR_BAR_CFGBITS_MASK = 0xF

PCI_HDR_BAR_IOMMIO_MASK = 0x1
PCI_HDR_BAR_IOMMIO_MMIO = 0
PCI_HDR_BAR_IOMMIO_IO = 1

PCI_HDR_BAR_TYPE_MASK = (0x3 << 1)
PCI_HDR_BAR_TYPE_SHIFT = 1
PCI_HDR_BAR_TYPE_64B = 2
PCI_HDR_BAR_TYPE_1MB = 1
PCI_HDR_BAR_TYPE_32B = 0

PCI_HDR_BAR_BASE_MASK_MMIO64 = 0xFFFFFFFFFFFFFFF0
PCI_HDR_BAR_BASE_MASK_MMIO = 0xFFFFFFF0
PCI_HDR_BAR_BASE_MASK_IO = 0xFFFC

# Type 0 specific registers
PCI_HDR_TYPE0_BAR1_LO_OFF = 0x18
PCI_HDR_TYPE0_BAR1_HI_OFF = 0x1C
PCI_HDR_TYPE0_BAR2_LO_OFF = 0x20
PCI_HDR_TYPE0_BAR2_HI_OFF = 0x24
PCI_HDR_TYPE0_XROM_BAR_OFF = 0x30

# Type 1 specific registers
PCI_HDR_TYPE1_XROM_BAR_OFF = 0x38

# Field defines

PCI_HDR_CMD_MS_MASK = 0x2

PCI_HDR_TYPE_TYPE_MASK = 0x7F
PCI_HDR_TYPE_MF_MASK = 0x80

PCI_TYPE0 = 0x0
PCI_TYPE1 = 0x1

PCI_HDR_XROM_BAR_EN_MASK = 0x00000001
PCI_HDR_XROM_BAR_BASE_MASK = 0xFFFFF000

PCI_HDR_BAR_STEP = 0x4


#
# Generic/standard PCI Expansion (Option) ROM
#

XROM_SIGNATURE = 0xAA55
PCI_XROM_HEADER_FMT = '<H22sH'
PCI_XROM_HEADER_SIZE = struct.calcsize(PCI_XROM_HEADER_FMT)


class PCI_XROM_HEADER(namedtuple('PCI_XROM_HEADER', 'Signature ArchSpecific PCIROffset')):
    __slots__ = ()

    def __str__(self) -> str:
        return f"""
PCI XROM
-----------------------------------
Signature       : 0x{self.Signature:04X} (= 0xAA55)
ArchSpecific    : {self.ArchSpecific.encode('hex').upper()}
PCIR Offset     : 0x{self.PCIROffset:04X}
"""

# @TBD: PCI Data Structure

#
# EFI specific PCI Expansion (Option) ROM
#


EFI_XROM_SIGNATURE = 0x0EF1
EFI_XROM_HEADER_FMT = '<HHIHHHBHH'
EFI_XROM_HEADER_SIZE = struct.calcsize(EFI_XROM_HEADER_FMT)


class EFI_XROM_HEADER(namedtuple('EFI_XROM_HEADER', 'Signature InitSize EfiSignature EfiSubsystem EfiMachineType CompressType Reserved EfiImageHeaderOffset PCIROffset')):
    __slots__ = ()

    def __str__(self) -> str:
        return f"""
EFI PCI XROM
---------------------------------------
Signature           : 0x{self.Signature:04X} (= 0xAA55)
Init Size           : 0x{self.InitSize:04X} (x 512 B)
EFI Signature       : 0x{self.EfiSignature:08X} (= 0x0EF1)
EFI Subsystem       : 0x{self.EfiSubsystem:04X}
EFI Machine Type    : 0x{self.EfiMachineType:04X}
Compression Type    : 0x{self.CompressType:04X}
Reserved            : 0x{self.Reserved:02X}
EFI Image Hdr Offset: 0x{self.EfiImageHeaderOffset:04X}
PCIR Offset         : 0x{self.PCIROffset:04X}
"""

#
# Legacy PCI Expansion (Option) ROM
#


XROM_HEADER_FMT = '<HBI17sH'
XROM_HEADER_SIZE = struct.calcsize(XROM_HEADER_FMT)


class XROM_HEADER(namedtuple('XROM_HEADER', 'Signature InitSize InitEP Reserved PCIROffset')):
    __slots__ = ()

    def __str__(self) -> str:
        return f"""
XROM
--------------------------------------
Signature           : 0x{self.Signature:04X}
Init Size           : 0x{self.InitSize:02X} (x 512 B)
Init Entry-point    : 0x{self.InitEP:08X}
Reserved            : {self.Reserved.encode('hex').upper()}
PCIR Offset         : 0x{self.PCIROffset:04X}
"""


class XROM:
    def __init__(self, bus, dev, fun, en, base, size):
        self.bus: int = bus
        self.dev: int = dev
        self.fun: int = fun
        self.vid: int = 0xFFFF
        self.did: int = 0xFFFF
        self.en: int = en
        self.base: int = base
        self.size: int = size
        self.header: Optional[PCI_XROM_HEADER] = None


def get_vendor_name_by_vid(vid: int) -> str:
    if vid in VENDORS:
        return VENDORS[vid]
    return ''


def get_device_name_by_didvid(vid: int, did: int) -> str:
    if vid in DEVICES:
        if did in DEVICES[vid]:
            return DEVICES[vid][did]
    return ''


def print_pci_devices(_devices: List[Tuple[int, int, int, int, int, int]]) -> None:
    logger().log("BDF     | VID:DID   | Vendor                       | Device")
    logger().log("-------------------------------------------------------------------------")
    for (b, d, f, vid, did, _) in _devices:
        vendor_name = get_vendor_name_by_vid(vid)
        device_name = get_device_name_by_didvid(vid, did)
        logger().log(f'{b:02X}:{d:02X}.{f:X} | {vid:04X}:{did:04X} | {vendor_name:28} | {device_name}')


def print_pci_XROMs(_xroms: List[XROM]) -> None:
    if len(_xroms) == 0:
        return None
    logger().log("BDF     | VID:DID   | XROM base | XROM size | en ")
    logger().log("-------------------------------------------------")
    for xrom in _xroms:
        logger().log(f'{xrom.bus:02X}:{xrom.dev:02X}.{xrom.fun:X} | {xrom.vid:04X}:{xrom.did:04X} | {xrom.base:08X}  | {xrom.size:08X}  | {xrom.en:d}')


class Pci:

    def __init__(self, cs):
        self.cs = cs
        self.helper = cs.helper
        self.hal_log_every_read = True

    #
    # Access to PCI configuration registers
    #

    def read_dword(self, bus: int, device: int, function: int, address: int) -> int:
        value = self.helper.read_pci_reg(bus, device, function, address, 4)
        if self.hal_log_every_read or value != 0xFFFFFFFF:
            logger().log_hal(f'[pci] reading B/D/F: {bus:d}/{device:d}/{function:d}, offset: 0x{address:02X}, value: 0x{value:08X}')
        return value

    def read_word(self, bus: int, device: int, function: int, address: int) -> int:
        word_value = self.helper.read_pci_reg(bus, device, function, address, 2)
        if self.hal_log_every_read or word_value != 0xFFFF:
            logger().log_hal(f'[pci] reading B/D/F: {bus:d}/{device:d}/{function:d}, offset: 0x{address:02X}, value: 0x{word_value:04X}')
        return word_value

    def read_byte(self, bus: int, device: int, function: int, address: int) -> int:
        byte_value = self.helper.read_pci_reg(bus, device, function, address, 1)
        if self.hal_log_every_read or byte_value != 0xFF:
            logger().log_hal(f'[pci] reading B/D/F: {bus:d}/{device:d}/{function:d}, offset: 0x{address:02X}, value: 0x{byte_value:02X}')
        return byte_value

    def write_byte(self, bus: int, device: int, function: int, address: int, byte_value: int) -> None:
        self.helper.write_pci_reg(bus, device, function, address, byte_value, 1)
        logger().log_hal(f'[pci] writing B/D/F: {bus:d}/{device:d}/{function:d}, offset: 0x{address:02X}, value: 0x{byte_value:02X}')
        return None

    def write_word(self, bus: int, device: int, function: int, address: int, word_value: int) -> None:
        self.helper.write_pci_reg(bus, device, function, address, word_value, 2)
        logger().log_hal(f'[pci] writing B/D/F: {bus:d}/{device:d}/{function:d}, offset: 0x{address:02X}, value: 0x{word_value:04X}')
        return None

    def write_dword(self, bus: int, device: int, function: int, address: int, dword_value: int) -> None:
        self.helper.write_pci_reg(bus, device, function, address, dword_value, 4)
        logger().log_hal(f'[pci] writing B/D/F: {bus:d}/{device:d}/{function:d}, offset: 0x{address:02X}, value: 0x{dword_value:08X}')
        return None

    #
    # Enumerating PCI devices and dumping configuration space
    #

    def enumerate_devices(self, bus: Optional[int] = None, device: Optional[int] = None, function: Optional[int] = None, spec: Optional[bool] = True) -> List[Tuple[int, int, int, int, int, int]]:
        devices = []
        self.hal_log_every_read = False
        if bus is not None:
            bus_range = [bus]
        else:
            bus_range = range(256)
        if device is not None:
            dev_range = [device]
        else:
            dev_range = range(32)
        if function is not None:
            func_range = [function]
        else:
            func_range = range(8)

        for b, d in itertools.product(bus_range, dev_range):
            for f in func_range:
                try:
                    did_vid = self.read_dword(b, d, f, 0x0)
                    if 0xFFFFFFFF != did_vid:
                        vid = did_vid & 0xFFFF
                        did = (did_vid >> 16) & 0xFFFF
                        rid = self.read_byte(b, d, f, 0x8)
                        devices.append((b, d, f, vid, did, rid))
                    elif f == 0 and spec:
                        break
                except OsHelperError:
                    logger().log_hal(f"[pci] unable to access B/D/F: {b:d}/{d:d}/{f:d}")
        self.hal_log_every_read = True
        return devices

    def dump_pci_config(self, bus: int, device: int, function: int) -> List[int]:
        cfg = []
        for off in range(0, 0x100, 4):
            tmp_val = self.read_dword(bus, device, function, off)
            for shift in range(0, 32, 8):
                cfg.append((tmp_val >> shift) & 0xFF)
        return cfg

    def print_pci_config_all(self) -> None:
        logger().log("[pci] enumerating available PCI devices...")
        pci_devices = self.enumerate_devices()
        for (b, d, f, vid, did, rid) in pci_devices:
            cfg_buf = self.dump_pci_config(b, d, f)
            logger().log(f"\n[pci] PCI device {b:02X}:{d:02X}.{f:02X} configuration:")
            pretty_print_hex_buffer(cfg_buf)

    #
    # PCI Expansion ROM functions
    #

    def parse_XROM(self, xrom: XROM, xrom_dump: bool = False) -> Optional[PCI_XROM_HEADER]:
        xrom_sig = self.cs.mem.read_physical_mem_word(xrom.base)
        if xrom_sig != XROM_SIGNATURE:
            return None
        xrom_hdr_buf = self.cs.mem.read_physical_mem(xrom.base, PCI_XROM_HEADER_SIZE)
        xrom_hdr = PCI_XROM_HEADER(*struct.unpack_from(PCI_XROM_HEADER_FMT, xrom_hdr_buf))
        if xrom_dump:
            xrom_fname = f'xrom_{xrom.bus:X}-{xrom.dev:X}-{xrom.fun:X}_{xrom.vid:X}{xrom.did:X}.bin'
            xrom_buf = self.cs.mem.read_physical_mem(xrom.base, xrom.size)  # use xrom_hdr.InitSize ?
            write_file(xrom_fname, xrom_buf)
        return xrom_hdr

    def find_XROM(self, bus: int, dev: int, fun: int, try_init: bool = False, xrom_dump: bool = False, xrom_addr: Optional[int] = None) -> Tuple[bool, Optional[XROM]]:
        # return results
        xrom_found, xrom = False, None

        logger().log_hal(f'[pci] checking XROM in {bus:02X}:{dev:02X}.{fun:02X}')

        cmd = self.read_word(bus, dev, fun, PCI_HDR_CMD_OFF)
        ms = (cmd & PCI_HDR_CMD_MS_MASK) == PCI_HDR_CMD_MS_MASK
        logger().log_hal(f'[pci]   PCI CMD (memory space = {ms:d}): 0x{cmd:04X}')

        hdr_type = self.read_byte(bus, dev, fun, PCI_HDR_TYPE_OFF)
        _mf = hdr_type & PCI_HDR_TYPE_MF_MASK
        _type = hdr_type & PCI_HDR_TYPE_TYPE_MASK
        xrom_bar_off = PCI_HDR_TYPE1_XROM_BAR_OFF if _type == PCI_TYPE1 else PCI_HDR_TYPE0_XROM_BAR_OFF

        xrom_bar = self.read_dword(bus, dev, fun, xrom_bar_off)
        xrom_exists = (xrom_bar != 0)

        if xrom_exists:
            logger().log_hal(f'[pci]   device programmed XROM BAR: 0x{xrom_bar:08X}')
        else:
            logger().log_hal(f'[pci]   device did not program XROM BAR: 0x{xrom_bar:08X}')
            if try_init:
                self.write_dword(bus, dev, fun, xrom_bar_off, PCI_HDR_XROM_BAR_BASE_MASK)
                xrom_bar = self.read_dword(bus, dev, fun, xrom_bar_off)
                xrom_exists = (xrom_bar != 0)
                logger().log_hal(f'[pci]   returned 0x{xrom_bar:08X} after writing {PCI_HDR_XROM_BAR_BASE_MASK:08X}')
                if xrom_exists and (xrom_addr is not None):
                    # device indicates XROM may exist. Initialize its base with supplied MMIO address
                    size_align = ~(xrom_bar & PCI_HDR_XROM_BAR_BASE_MASK)  # actual XROM alignment
                    if (xrom_addr & size_align) != 0:
                        logger().log_warning(f'XROM address 0x{xrom_addr:08X} must be aligned at 0x{size_align:08X}')
                        return False, None
                    self.write_dword(bus, dev, fun, xrom_bar_off, (xrom_addr | PCI_HDR_XROM_BAR_EN_MASK))
                    xrom_bar = self.read_dword(bus, dev, fun, xrom_bar_off)
                    logger().log_hal(f'[pci]   programmed XROM BAR with 0x{xrom_bar:08X}')

        #
        # At this point, a device indicates that XROM exists. Let's check if XROM is really there
        #
        xrom_en = (xrom_bar & PCI_HDR_XROM_BAR_EN_MASK) == 0x1
        xrom_base = xrom_bar & PCI_HDR_XROM_BAR_BASE_MASK
        xrom_size = ~xrom_base + 1

        if xrom_exists:
            logger().log_hal(f'[pci]   XROM: BAR = 0x{xrom_bar:08X}, base = 0x{xrom_base:08X}, size = 0x{xrom_size:X}, en = {xrom_en:d}')
            xrom = XROM(bus, dev, fun, xrom_en, xrom_base, xrom_size)
            if xrom_en and (xrom_base != PCI_HDR_XROM_BAR_BASE_MASK):
                xrom.header = self.parse_XROM(xrom, xrom_dump)
                xrom_found = (xrom is not None) and (xrom.header is not None)
                if xrom_found:
                    logger().log_hal(f"[pci]   XROM found at 0x{xrom_base:08X}")
                    logger().log_hal(str(xrom.header))

        if not xrom_found:
            logger().log_hal('[pci]   XROM was not found')

        return xrom_found, xrom

    def enumerate_xroms(self, try_init: bool = False, xrom_dump: bool = False, xrom_addr: Optional[int] = None) -> List[Optional[XROM]]:
        pci_xroms = []
        logger().log("[pci] enumerating available PCI devices...")
        pci_devices = self.enumerate_devices()
        for (b, d, f, vid, did, rid) in pci_devices:
            exists, xrom = self.find_XROM(b, d, f, try_init, xrom_dump, xrom_addr)
            if exists and (xrom is not None):
                xrom.vid = vid
                xrom.did = did
                pci_xroms.append(xrom)
        return pci_xroms

    def get_header_type(self, bus, dev, fun):
        res = self.read_byte(bus, dev, fun, PCI_HDR_TYPE_OFF)
        return res & PCI_HDR_TYPE_TYPE_MASK

    #
    # Calculates actual size of MMIO BAR range
    def calc_bar_size(self, bus: int, dev: int, fun: int, off: int, is64: bool, isMMIO: bool) -> int:
        logger().log_hal(f'calc_bar_size {bus}:{dev}.{fun} offset{off}')
        # Read the original value of the register
        orig_regL = self.read_dword(bus, dev, fun, off)
        logger().log_hal(f'orig_regL: {orig_regL:X}')
        if is64:
            orig_regH = self.read_dword(bus, dev, fun, off + PCI_HDR_BAR_STEP)
            logger().log_hal(f'orig_regH: {orig_regH:X}')
        # Write all 1's to the register
        self.write_dword(bus, dev, fun, off + PCI_HDR_BAR_STEP, MASK_32b)
        if is64:
            self.write_dword(bus, dev, fun, off, MASK_32b)
        # Read the register back
        regL = self.read_dword(bus, dev, fun, off)
        logger().log_hal(f'regL: {regL:X}')
        if is64:
            regH = self.read_dword(bus, dev, fun, off + PCI_HDR_BAR_STEP)
            logger().log_hal(f'regH: {regH:X}')
        # Write original value back to register
        self.write_dword(bus, dev, fun, off, orig_regL)
        if is64:
            self.write_dword(bus, dev, fun, off + PCI_HDR_BAR_STEP, orig_regH)
        # Calculate Sizing
        if isMMIO and is64:
            reg = regL | (regH << 32)
            orig_reg = orig_regL | (orig_regH << 32)
            if orig_reg == reg:
                size = BOUNDARY_4KB
            else:
                size = (~(reg & PCI_HDR_BAR_BASE_MASK_MMIO64) & MASK_64b) + 1
        elif isMMIO:
            if regL == orig_regL:
                size = BOUNDARY_4KB
            else:
                size = (~(regL & PCI_HDR_BAR_BASE_MASK_MMIO) & MASK_32b) + 1
        else:
            if regL == orig_regL:
                size = 0x100
            else:
                size = (~(regL & PCI_HDR_BAR_BASE_MASK_IO) & MASK_16b) + 1
        return size

    # Returns all I/O and MMIO BARs defined in the PCIe header of the device
    # Returns array of elements in format (BAR_address, isMMIO, is64bit, BAR_reg_offset, BAR_reg_value)
    def get_device_bars(self, bus: int, dev: int, fun: int, bCalcSize: bool = False) -> List[Tuple[int, bool, bool, int, int, int]]:
        _bars = []
        hdr_type = self.get_header_type(bus, dev, fun)
        if hdr_type == 0:
            bounds = PCI_HDR_TYPE0_BAR2_HI_OFF
        elif hdr_type == 1:
            bounds = PCI_HDR_TYPE0_BAR1_LO_OFF
        else:
            bounds = PCI_HDR_BAR0_LO_OFF

        off = PCI_HDR_BAR0_LO_OFF
        size = BOUNDARY_4KB
        while off <= bounds:
            reg = self.read_dword(bus, dev, fun, off)
            if reg and reg != MASK_32b:
                # BAR is initialized
                isMMIO = (PCI_HDR_BAR_IOMMIO_MMIO == (reg & PCI_HDR_BAR_IOMMIO_MASK))
                if isMMIO:
                    # MMIO BAR
                    mem_type = (reg & PCI_HDR_BAR_TYPE_MASK) >> PCI_HDR_BAR_TYPE_SHIFT
                    if PCI_HDR_BAR_TYPE_64B == mem_type:
                        # 64-bit MMIO BAR
                        if bCalcSize and hdr_type == 0:
                            size = self.calc_bar_size(bus, dev, fun, off, True, True)
                        off += PCI_HDR_BAR_STEP
                        reg_hi = self.read_dword(bus, dev, fun, off)
                        reg |= (reg_hi << 32)
                        base = (reg & PCI_HDR_BAR_BASE_MASK_MMIO64)
                        if base != 0:
                            _bars.append((base, isMMIO, True, off - PCI_HDR_BAR_STEP, reg, size))
                    elif PCI_HDR_BAR_TYPE_1MB == mem_type:
                        # MMIO BAR below 1MB - not supported
                        pass
                    elif PCI_HDR_BAR_TYPE_32B == mem_type:
                        # 32-bit only MMIO BAR
                        base = (reg & PCI_HDR_BAR_BASE_MASK_MMIO)
                        if base != 0:
                            if bCalcSize and hdr_type == 0:
                                size = self.calc_bar_size(bus, dev, fun, off, False, True)
                            _bars.append((base, isMMIO, False, off, reg, size))
                else:
                    # I/O BAR
                    base = (reg & PCI_HDR_BAR_BASE_MASK_IO)
                    if base != 0:
                        if bCalcSize and hdr_type == 0:
                            size = self.calc_bar_size(bus, dev, fun, off, False, False)
                        else:
                            size = 0x100
                        _bars.append((base, isMMIO, False, off, reg, size))
            off += PCI_HDR_BAR_STEP
        return _bars

    def get_DIDVID(self, bus: int, dev: int, fun: int) -> Tuple[int, int]:
        didvid = self.read_dword(bus, dev, fun, 0x0)
        vid = didvid & 0xFFFF
        did = (didvid >> 16) & 0xFFFF
        return (did, vid)

    def is_enabled(self, bus: int, dev: int, fun: int) -> bool:
        (did, vid) = self.get_DIDVID(bus, dev, fun)
        if (is_all_ones(vid, 2)) or (is_all_ones(did, 2)):
            return False
        return True



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
The msr command allows direct access to read and write MSRs.

>>> chipsec_util msr <msr> [eax] [edx] [thread_id]

Examples:

>>> chipsec_util msr 0x3A
>>> chipsec_util msr 0x3A 0x0
>>> chipsec_util msr 0x8B 0x0 0x0 0x0
"""

from chipsec.command import BaseCommand, toLoad
from argparse import ArgumentParser


# CPU Model Specific Registers
class MSRCommand(BaseCommand):

    def requirements(self) -> toLoad:
        return toLoad.Driver

    def parse_arguments(self) -> None:
        parser = ArgumentParser(prog='chipsec_util msr', usage=__doc__)
        parser.add_argument('msr_addr', type=lambda x: int(x, 0), metavar='<msr>', help='MSR address')
        parser.add_argument('msr_input1', type=lambda x: int(x, 0), metavar='MSR Value', nargs='?', default=None, help='EAX (Low)')
        parser.add_argument('msr_input2', type=lambda x: int(x, 0), metavar='MSR Value', nargs='?', default=None, help='EDX (High)')
        parser.add_argument('thread_id', type=lambda x: int(x, 0), metavar='Thread ID', nargs='?', default=None, help='Thread ID')
        parser.parse_args(self.argv, namespace=self)

    def run(self):
        if self.msr_input1 is None:
            for tid in range(self.cs.msr.get_cpu_thread_count()):
                (eax, edx) = self.cs.msr.read_msr(tid, self.msr_addr)
                val64 = ((edx << 32) | eax)
                self.logger.log("[CHIPSEC] CPU{:d}: RDMSR( 0x{:x} ) = {:016X} (EAX={:08X}, EDX={:08X})".format(tid, self.msr_addr, val64, eax, edx))
        elif self.msr_input2 is None:
            cpu_thread_id = self.msr_input1
            (eax, edx) = self.cs.msr.read_msr(cpu_thread_id, self.msr_addr)
            val64 = ((edx << 32) | eax)
            self.logger.log("[CHIPSEC] CPU{:d}: RDMSR( 0x{:x} ) = {:016X} (EAX={:08X}, EDX={:08X})".format(cpu_thread_id, self.msr_addr, val64, eax, edx))
        else:
            eax = self.msr_input1
            edx = self.msr_input2
            val64 = ((edx << 32) | eax)
            if self.thread_id is None:
                self.logger.log("[CHIPSEC] All CPUs: WRMSR( 0x{:x} ) = {:016X}".format(self.msr_addr, val64))
                for tid in range(self.cs.msr.get_cpu_thread_count()):
                    self.cs.msr.write_msr(tid, self.msr_addr, eax, edx)
            else:
                cpu_thread_id = self.thread_id
                self.logger.log("[CHIPSEC] CPU{:d}: WRMSR( 0x{:x} ) = {:016X}".format(cpu_thread_id, self.msr_addr, val64))
                self.cs.msr.write_msr(cpu_thread_id, self.msr_addr, eax, edx)


commands = {'msr': MSRCommand}



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
>>> chipsec_util check list
>>> chipsec_util check lock <lockname>
>>> chipsec_util check lock <lockname1, lockname2, ...>
>>> chipsec_util check all

Examples:

>>> chipsec_util check list
>>> chipsec_util check lock DebugLock
>>> chipsec_util check all

KEY:
    Lock Name - Name of Lock within configuration file
    State     - Lock Configuration

        Undefined - Lock is not defined within configuration
        Undoc     - Lock is missing configuration information
        Hidden    - Lock is in a disabled or hidden state (unable to read the lock)
        Unlocked  - Lock does not match value within configuration
        Locked    - Lock matches value within configuration
        RW/O      - Lock is identified as register is RW/O

"""

from argparse import ArgumentParser

from chipsec.command import BaseCommand, toLoad
from chipsec.hal.locks import locks, LockResult
from chipsec.library.defines import is_set


class LOCKCHECKCommand(BaseCommand):

    version = "0.5"

    def requirements(self) -> toLoad:
        return toLoad.All

    def parse_arguments(self) -> None:
        parser = ArgumentParser(prog='chipsec_util check', usage=LOCKCHECKCommand.__doc__)

        parser_lockname = ArgumentParser(add_help=False)
        parser_lockname.add_argument('lockname', type=str, nargs='+', help="locknames")

        subparsers = parser.add_subparsers()

        # list
        parser_list = subparsers.add_parser('list')
        parser_list.set_defaults(func=self.list_locks)

        # checkall
        parser_checkall = subparsers.add_parser('all')
        parser_checkall.set_defaults(func=self.checkall_locks)

        # check
        parser_check = subparsers.add_parser('lock', parents=[parser_lockname])
        parser_check.set_defaults(func=self.check_lock)

        parser.parse_args(self.argv, namespace=self)

    def set_up(self) -> None:
        self.flip_consistency_checking = False
        if not self.cs.consistency_checking:
            self.flip_consistency_checking = True
            self.cs.consistency_checking = True
        self.logger.set_always_flush(True)
        self._locks = locks(self.cs)

    def tear_down(self) -> None:
        self.logger.set_always_flush(False)
        if self.flip_consistency_checking:
            self.cs.consistency_checking = False

    def log_key(self) -> None:
        self.logger.log("""
KEY:
\tLock Name - Name of Lock within configuration file
\tState - Lock Configuration
\t\tUndefined - Lock is not defined within configuration
\t\tUndoc - Lock is missing configuration information
\t\tHidden - Lock is in a disabled or hidden state (unable to read the lock)
\t\tUnlocked - Lock does not match value within configuration
\t\tLocked - Lock matches value within configuration
\t\tRW/O - Lock is identified as register is RW/O\n\n""")

    def log_header(self) -> str:
        ret = f'{"Lock Name":^27}|{"State":^16}|{"Consistent":^16}\n{"-" * 58}'
        if not self.logger.HAL:
            self.logger.log(ret)
        return f"\n\n{ret}"

    def list_locks(self) -> None:
        self.logger.log('Locks identified within the configuration:')
        for lock in self._locks.get_locks():
            self.logger.log(lock)
        self.logger.log('')
        return

    def checkall_locks(self) -> None:
        locks = self._locks.get_locks()
        if not locks:
            self.logger.log('Did not find any locks')
            return
        if self.logger.VERBOSE:
            self.log_key()
        res = self.log_header()
        for lock in locks:
            is_locked = self._locks.is_locked(lock)
            is_locked_str = self.check_log(lock, is_locked)
            res = f"{res}\n{is_locked_str}"
        if self.logger.HAL:
            self.logger.log(res)
        return

    def check_lock(self) -> None:
        if self.logger.VERBOSE:
            self.log_key()
        res = self.log_header()
        for lock in self.lockname:
            is_locked = self._locks.is_locked(lock)
            is_locked_str = self.check_log(lock, is_locked)
            res = f"{res}\n{is_locked_str}"
        if self.logger.HAL:
            self.logger.log(res)
        return

    def check_log(self, lock: str, is_locked: int) -> str:
        consistent = "N/A"
        if not is_set(is_locked, LockResult.DEFINED):
            res_str = 'Undefined'
        elif not is_set(is_locked, LockResult.HAS_CONFIG):
            res_str = 'Undoc'
        elif not is_set(is_locked, LockResult.CAN_READ):
            res_str = 'Hidden'
        elif self.cs.lock.get_type(lock) == "RW/O":
            res_str = 'RW/O'
        elif is_set(is_locked, LockResult.LOCKED):
            res_str = 'Locked'
        elif not is_set(is_locked, LockResult.LOCKED):
            res_str = 'UnLocked'
        else:
            res_str = 'Unknown'
        if res_str in ["RW/O", "Locked", "UnLocked"] and is_set(is_locked, LockResult.INCONSISTENT):
            consistent = "No"
        elif res_str in ["RW/O", "Locked", "UnLocked"] and not is_set(is_locked, LockResult.INCONSISTENT):
            consistent = "Yes"
        res = f'{lock[:26]:27}|  {res_str:14}|{consistent:^16}'
        if not self.logger.HAL:
            self.logger.log(res)
        return res


commands = {'check': LOCKCHECKCommand}



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
Functionality encapsulating interrupt generation
CPU Interrupts specific functions (SMI, NMI)

usage:
    >>> send_SMI_APMC( 0xDE )
    >>> send_NMI()
"""

# TODO IPIs through Local APIC??

import struct
import uuid
from typing import Optional, Tuple
from chipsec.hal import hal_base
from chipsec.library.logger import logger, print_buffer_bytes
from chipsec.hal.acpi import ACPI
from chipsec.hal.acpi_tables import UEFI_TABLE, GAS
from chipsec.library.defines import bytestostring

SMI_APMC_PORT = 0xB2
SMI_DATA_PORT = 0xB3

NMI_TCO1_CTL = 0x8  # NMI_NOW is bit [8] in TCO1_CTL (or bit [1] in TCO1_CTL + 1)
NMI_NOW = 0x1


class Interrupts(hal_base.HALBase):

    def __init__(self, cs):
        super(Interrupts, self).__init__(cs)

    def send_SW_SMI(self, thread_id: int, SMI_code_port_value: int, SMI_data_port_value: int, _rax: int, _rbx: int, _rcx: int, _rdx: int, _rsi: int, _rdi: int) -> Optional[Tuple[int, int, int, int, int, int, int]]:
        SMI_code_data = (SMI_data_port_value << 8 | SMI_code_port_value)
        logger().log_hal(
            f"[intr] Sending SW SMI: code port 0x{SMI_APMC_PORT:02X} <- 0x{SMI_code_port_value:02X}, data port 0x{SMI_APMC_PORT + 1:02X} <- 0x{SMI_data_port_value:02X} (0x{SMI_code_data:04X})")
        logger().log_hal(f"       RAX = 0x{_rax:016X} (AX will be overridden with values of SW SMI ports B2/B3)")
        logger().log_hal(f"       RBX = 0x{_rbx:016X}")
        logger().log_hal(f"       RCX = 0x{_rcx:016X}")
        logger().log_hal(f"       RDX = 0x{_rdx:016X} (DX will be overridden with 0x00B2)")
        logger().log_hal(f"       RSI = 0x{_rsi:016X}")
        logger().log_hal(f"       RDI = 0x{_rdi:016X}")
        return self.cs.helper.send_sw_smi(thread_id, SMI_code_data, _rax, _rbx, _rcx, _rdx, _rsi, _rdi)

    def send_SW_SMI_timed(self, thread_id: int, SMI_code_port_value: int, SMI_data_port_value: int, _rax: int, _rbx: int, _rcx: int, _rdx: int, _rsi: int, _rdi: int) -> Optional[Tuple[int, int, int, int, int, int, int]]:
        SMI_code_data = (SMI_data_port_value << 8 | SMI_code_port_value)
        logger().log_hal(
            f"[intr] Sending SW SMI: code port 0x{SMI_APMC_PORT:02X} <- 0x{SMI_code_port_value:02X}, data port 0x{SMI_APMC_PORT + 1:02X} <- 0x{SMI_data_port_value:02X} (0x{SMI_code_data:04X})")
        logger().log_hal(f"       RAX = 0x{_rax:016X} (AX will be overridden with values of SW SMI ports B2/B3)")
        logger().log_hal(f"       RBX = 0x{_rbx:016X}")
        logger().log_hal(f"       RCX = 0x{_rcx:016X}")
        logger().log_hal(f"       RDX = 0x{_rdx:016X} (DX will be overridden with 0x00B2)")
        logger().log_hal(f"       RSI = 0x{_rsi:016X}")
        logger().log_hal(f"       RDI = 0x{_rdi:016X}")
        return self.cs.helper.send_sw_smi_timed(thread_id, SMI_code_data, _rax, _rbx, _rcx, _rdx, _rsi, _rdi)

    def send_SMI_APMC(self, SMI_code_port_value: int, SMI_data_port_value: int) -> None:
        logger().log_hal(f"[intr] sending SMI via APMC ports: code 0xB2 <- 0x{SMI_code_port_value:02X}, data 0xB3 <- 0x{SMI_data_port_value:02X}")
        self.cs.io.write_port_byte(SMI_DATA_PORT, SMI_data_port_value)
        return self.cs.io.write_port_byte(SMI_APMC_PORT, SMI_code_port_value)

    def send_NMI(self) -> None:
        logger().log_hal("[intr] Sending NMI# through TCO1_CTL[NMI_NOW]")
        reg, ba = self.cs.device.get_IO_space("TCOBASE")
        tcobase = self.cs.register.read_field(reg, ba)
        return self.cs.io.write_port_byte(tcobase + NMI_TCO1_CTL + 1, NMI_NOW)

    def find_ACPI_SMI_Buffer(self) -> Optional[UEFI_TABLE.CommBuffInfo]:
        logger().log_hal("Parsing ACPI tables to identify Communication Buffer")
        _acpi = ACPI(self.cs).get_ACPI_table("UEFI")
        if len(_acpi):
            _uefi = UEFI_TABLE()
            _uefi.parse(_acpi[0][1])
            logger().log_hal(str(_uefi))
            return _uefi.get_commbuf_info()
        logger().log_hal("Unable to find Communication Buffer")
        return None

    def send_ACPI_SMI(self, thread_id: int, smi_num: int, buf_addr: int, invoc_reg: GAS, guid: str, data: bytes) -> Optional[int]:
        # Prepare Communication Data buffer
        # typedef struct {
        #   EFI_GUID HeaderGuid;
        #   UINTN MessageLength;
        #   UINT8 Data[ANYSIZE_ARRAY];
        # } EFI_SMM_COMMUNICATE_HEADER;
        _guid = uuid.UUID(guid).bytes_le
        data_hdr = _guid + struct.pack("Q", len(data)) + data
        if not invoc_reg is None:
            # need to write data_hdr to comm buffer
            self.cs.helper.write_phys_mem(buf_addr, len(data_hdr), data_hdr)
            # USING GAS need to write buf_addr into invoc_reg
            if invoc_reg.addrSpaceID == 0:
                self.cs.helper.write_phys_mem(invoc_reg.addr, invoc_reg.accessSize, buf_addr)
                # check for return status
                ret_buf = self.cs.helper.read_phys_mem(buf_addr, 8)
            elif invoc_reg.addrSpaceID == 1:
                self.cs.helper.write_io_port(invoc_reg.addr, invoc_reg.accessSize, buf_addr)
                # check for return status
                ret_buf = self.cs.helper.read_io_port(buf_addr, 8)
            else:
                logger().log_error("Functionality is currently not implemented")
                ret_buf = None
            return ret_buf

        else:
            # Wait for Communication buffer to be empty
            buf = 1
            while not buf == b"\x00\x00":
                buf = self.cs.helper.read_phys_mem(buf_addr, 2)
            # write data to commbuffer
            self.cs.helper.write_phys_mem(buf_addr, len(data_hdr), data_hdr)
            # call SWSMI
            self.send_SW_SMI(thread_id, smi_num, 0, 0, 0, 0, 0, 0, 0)
            # clear CommBuffer
            self.cs.helper.write_phys_mem(buf_addr, len(data_hdr), b"\x00" * len(data_hdr))
            return None

    # scan phys mem range start-end looking for 'smmc'
    def find_smmc(self, start: int, end: int) -> int:
        chunk_sz = 1024 * 8  # 8KB chunks
        phys_address = start
        found_at = 0
        while phys_address <= end:
            buffer = self.cs.mem.read_physical_mem(phys_address, chunk_sz)
            buffer = bytestostring(buffer)
            offset = buffer.find('smmc')
            if offset != -1:
                found_at = phys_address + offset
                break
            phys_address += chunk_sz
        return found_at

    '''
Send SWSMI in the same way as EFI_SMM_COMMUNICATION_PROTOCOL
    - Write Commbuffer location and Commbuffer size to 'smmc' structure
    - Write 0 to 0xb3 and 0xb2

MdeModulePkg/Core/PiSmmCore/PiSmmCorePrivateData.h

#define SMM_CORE_PRIVATE_DATA_SIGNATURE  SIGNATURE_32 ('s', 'm', 'm', 'c')
 struct {
  UINTN                           Signature;
   This field is used by the SMM Communicatioon Protocol to pass a buffer into
   a software SMI handler and for the software SMI handler to pass a buffer back to
   the caller of the SMM Communication Protocol.
  VOID                            *CommunicationBuffer;
  UINTN                           BufferSize;

  EFI_STATUS                      ReturnStatus;
} SMM_CORE_PRIVATE_DATA;
    '''

    def send_smmc_SMI(self, smmc: int, guid: str, payload: bytes, payload_loc: int, CommandPort: int = 0x0, DataPort: int = 0x0) -> int:
        guid_b = uuid.UUID(guid).bytes_le
        payload_sz = len(payload)

        data_hdr = guid_b + struct.pack("Q", payload_sz) + payload
        # write payload to payload_loc
        CommBuffer_offset = 56
        BufferSize_offset = CommBuffer_offset + 8
        ReturnStatus_offset = BufferSize_offset + 8

        self.cs.mem.write_physical_mem(smmc + CommBuffer_offset, 8, struct.pack("Q", payload_loc))
        self.cs.mem.write_physical_mem(smmc + BufferSize_offset, 8, struct.pack("Q", len(data_hdr)))
        self.cs.mem.write_physical_mem(payload_loc, len(data_hdr), data_hdr)

        if self.logger.VERBOSE:
            self.logger.log("[*] Communication buffer on input")
            print_buffer_bytes(self.cs.mem.read_physical_mem(payload_loc, len(data_hdr)))
            self.logger.log("")

        self.send_SMI_APMC(CommandPort, DataPort)

        if self.logger.VERBOSE:
            self.logger.log("[*] Communication buffer on output")
            print_buffer_bytes(self.cs.mem.read_physical_mem(payload_loc, len(data_hdr)))
            self.logger.log("")

        ReturnStatus = struct.unpack("Q", self.cs.mem.read_physical_mem(smmc + ReturnStatus_offset, 8))[0]
        return ReturnStatus



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


#
# -------------------------------------------------------------------------------
#
# CHIPSEC: Platform Hardware Security Assessment Framework
# (c) 2010-2012 Intel Corporation
#
# -------------------------------------------------------------------------------

"""
Reading from/writing to files

usage:
    >>> read_file(filename)
    >>> write_file(filename, buffer)
"""

import os
from typing import Any
from chipsec.library.strings import get_datetime_str
from chipsec.library.logger import logger

TOOLS_DIR = 'chipsec_tools'


def read_file(filename: str, size: int = 0) -> bytes:
    try:
        with open(filename, 'rb') as f:
            if size:
                _file = f.read(size)
            else:
                _file = f.read()
            logger().log_debug(f"[file] Read {len(_file):d} bytes from '{filename:.256}'")
            return _file
    except OSError:
        logger().log_error(f"Unable to open file '{filename:.256}' for read access")
        return b''


def write_file(filename: str, buffer: Any, append: bool = False) -> bool:
    perm = 'a' if append else 'w'
    if isinstance(buffer, bytes) or isinstance(buffer, bytearray):
        perm += 'b'
    try:
        f = open(filename, perm)
    except OSError:
        logger().log_error(f"Unable to open file '{filename:.256}' for write access")
        return False
    f.write(buffer)
    f.close()

    logger().log_debug(f"[file] Wrote {len(buffer):d} bytes to '{filename:.256}'")
    return True


def write_unique_file(file_buffer: Any, file_name: str = '', file_extension: str = '') -> str:
    """Writes file with the name <file_name>_<year><month><day>-<hour><minute><second>.<file_extension>"""
    file_str = f'{file_name}_' if file_name else ''
    file_ext = f'.{file_extension}' if file_extension else ''
    file_name_str = f'{file_str}{get_datetime_str()}{file_ext}'
    return file_name_str if write_file(file_name_str, file_buffer) else ''


def get_main_dir() -> str:
    path = os.path.abspath(os.path.join(os.path.dirname(__file__), os.path.pardir, os.path.pardir))
    return path


def get_module_dir() -> str:
    path = os.path.join(get_main_dir(), "chipsec", "modules")
    return path



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2016, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
Xen specific defines
"""

hypercall_names = {
    0: 'set_trap_table',
    1: 'mmu_update',
    2: 'set_gdt',
    3: 'stack_switch',
    4: 'set_callbacks',
    5: 'fpu_taskswitch',
    6: 'sched_op_compat',
    7: 'platform_op',
    8: 'set_debugreg',
    9: 'get_debugreg',
    10: 'update_descriptor',
    12: 'memory_op',
    13: 'multicall',
    14: 'update_va_mapping',
    15: 'set_timer_op',
    16: 'event_channel_op_compat',
    17: 'xen_version',
    18: 'console_io',
    19: 'physdev_op_compat',
    20: 'grant_table_op',
    21: 'vm_assist',
    22: 'update_va_mapping_otherdomain',
    23: 'iret',
    24: 'vcpu_op',
    25: 'set_segment_base',
    26: 'mmuext_op',
    27: 'xsm_op',
    28: 'nmi_op',
    29: 'sched_op',
    30: 'callback_op',
    31: 'xenoprof_op',
    32: 'event_channel_op',
    33: 'physdev_op',
    34: 'hvm_op',
    35: 'sysctl',
    36: 'domctl',
    37: 'kexec_op',
    38: 'tmem_op',
    39: 'xc_reserved_op',
    40: 'xenpmu_op',
    48: 'arch_0',
    49: 'arch_1',
    50: 'arch_2',
    51: 'arch_3',
    52: 'arch_4',
    53: 'arch_5',
    54: 'arch_6',
    55: 'arch_7'
}


def get_hypercall_name(vector, defvalue=''):
    return hypercall_names[vector].upper() if vector in hypercall_names else defvalue


hypercall_status_codes = {
    0: ['XEN_STATUS_SUCCESS', 'Status success'],
    1: ['XEN_ERRNO_EPERM', 'Operation not permitted'],
    2: ['XEN_ERRNO_ENOENT', 'No such file or directory'],
    3: ['XEN_ERRNO_ESRCH', 'No such process'],
    4: ['XEN_ERRNO_EINTR', 'Interrupted system call'],
    5: ['XEN_ERRNO_EIO', 'I/O error'],
    6: ['XEN_ERRNO_ENXIO', 'No such device or address'],
    7: ['XEN_ERRNO_E2BIG', 'Arg list too long'],
    8: ['XEN_ERRNO_ENOEXEC', 'Exec format error'],
    9: ['XEN_ERRNO_EBADF', 'Bad file number'],
    10: ['XEN_ERRNO_ECHILD', 'No child processes'],
    11: ['XEN_ERRNO_EAGAIN', 'Try again'],
    12: ['XEN_ERRNO_ENOMEM', 'Out of memory'],
    13: ['XEN_ERRNO_EACCES', 'Permission denied'],
    14: ['XEN_ERRNO_EFAULT', 'Bad address'],
    16: ['XEN_ERRNO_EBUSY', 'Device or resource busy'],
    17: ['XEN_ERRNO_EEXIST', 'File exists'],
    18: ['XEN_ERRNO_EXDEV', 'Cross-device link'],
    19: ['XEN_ERRNO_ENODEV', 'No such device'],
    22: ['XEN_ERRNO_EINVAL', 'Invalid argument'],
    23: ['XEN_ERRNO_ENFILE', 'File table overflow'],
    24: ['XEN_ERRNO_EMFILE', 'Too many open files'],
    28: ['XEN_ERRNO_ENOSPC', 'No space left on device'],
    31: ['XEN_ERRNO_EMLINK', 'Too many links'],
    33: ['XEN_ERRNO_EDOM', 'Math argument out of domain of func'],
    34: ['XEN_ERRNO_ERANGE', 'Math result not representable'],
    35: ['XEN_ERRNO_EDEADLK', 'Resource deadlock would occur'],
    36: ['XEN_ERRNO_ENAMETOOLONG', 'File name too long'],
    37: ['XEN_ERRNO_ENOLCK', 'No record locks available'],
    38: ['XEN_ERRNO_ENOSYS', 'Function not implemented'],
    61: ['XEN_ERRNO_ENODATA', 'No data available'],
    62: ['XEN_ERRNO_ETIME', 'Timer expired'],
    74: ['XEN_ERRNO_EBADMSG', 'Not a data message'],
    75: ['XEN_ERRNO_EOVERFLOW', 'Value too large for defined data type'],
    84: ['XEN_ERRNO_EILSEQ', 'Illegal byte sequence'],
    85: ['XEN_ERRNO_ERESTART', 'Interrupted system call should be restarted'],
    88: ['XEN_ERRNO_ENOTSOCK', 'Socket operation on non-socket'],
    95: ['XEN_ERRNO_EOPNOTSUPP', 'Operation not supported on transport endpoint'],
    98: ['XEN_ERRNO_EADDRINUSE', 'Address already in use'],
    99: ['XEN_ERRNO_EADDRNOTAVAIL', 'Cannot assign requested address'],
    105: ['XEN_ERRNO_ENOBUFS', 'No buffer space available'],
    106: ['XEN_ERRNO_EISCONN', 'Transport endpoint is already connected'],
    107: ['XEN_ERRNO_ENOTCONN', 'Transport endpoint is not connected'],
    110: ['XEN_ERRNO_ETIMEDOUT', 'Connection timed out']
}


def get_iverr(status, bits=64):
    mask = (1 << bits) - 1
    return (mask - status + 1) & mask


def get_hypercall_status(code, brief=False):
    defstatus = [f'0x{code:016X}', f'Status 0x{code:016X}']
    if (code >> 32) == 0xFFFFFFFF:
        code = get_iverr(code)
        defstatus = [f'XEN_ERRNO_{code:04X}', f'Unknown error 0x{code:04X}']
    desc = hypercall_status_codes.get(code, defstatus)
    return desc[0] if brief else desc[1]


def get_hypercall_status_extended(code):
    return f'{get_hypercall_status(code, False)} - {get_hypercall_status(code, True)}'


def get_invalid_hypercall_code():
    return XEN_ERRNO_ENOSYS


xenmem_commands = {
    0: 'XENMEM_INCREASE_RESERVATION',
    1: 'XENMEM_DECREASE_RESERVATION',
    2: 'XENMEM_MAXIMUM_RAM_PAGE',
    3: 'XENMEM_CURRENT_RESERVATION',
    4: 'XENMEM_MAXIMUM_RESERVATION',
    5: 'XENMEM_MACHPHYS_MFN_LIST',
    6: 'XENMEM_POPULATE_PHYSMAP',
    7: 'XENMEM_ADD_TO_PHYSMAP',
    8: 'XENMEM_TRANSLATE_GPFN_LIST',
    9: 'XENMEM_MEMORY_MAP',
    10: 'XENMEM_MACHINE_MEMORY_MAP',
    11: 'XENMEM_EXCHANGE',
    12: 'XENMEM_MACHPHYS_MAPPING',
    15: 'XENMEM_REMOVE_FROM_PHYSMAP',
    23: 'XENMEM_ADD_TO_PHYSMAP_RANGE'
}

xen_version_commands = {
    0: 'XENVER_VERSION',
    1: 'XENVER_EXTRAVERSION',
    2: 'XENVER_COMPILE_INFO',
    3: 'XENVER_CAPABILITIES',
    4: 'XENVER_CHANGESET',
    5: 'XENVER_PLATFORM_PARAMETERS',
    6: 'XENVER_GET_FEATURES',
    7: 'XENVER_PAGESIZE',
    8: 'XENVER_GUEST_HANDLE',
    9: 'XENVER_COMMANDLINE'
}

console_io_commands = {
    0: 'CONSOLEIO_WRITE',
    1: 'CONSOLEIO_READ'
}

schedop_commands = {
    0: 'SCHEDOP_YIELD',
    1: 'SCHEDOP_BLOCK',
    2: 'SCHEDOP_SHUTDOWN',
    3: 'SCHEDOP_POLL',
    4: 'SCHEDOP_REMOTE_SHUTDOWN',
    5: 'SCHEDOP_SHUTDOWN_CODE',
    6: 'SCHEDOP_WATCHDOG'
}

evtchop_commands = {
    0: 'EVTCHOP_BIND_INTERDOMAIN',
    1: 'EVTCHOP_BIND_VIRQ',
    2: 'EVTCHOP_BIND_PIRQ',
    3: 'EVTCHOP_CLOSE',
    4: 'EVTCHOP_SEND',
    5: 'EVTCHOP_STATUS',
    6: 'EVTCHOP_ALLOC_UNBOUND',
    7: 'EVTCHOP_BIND_IPI',
    8: 'EVTCHOP_BIND_VCPU',
    9: 'EVTCHOP_UNMASK',
    10: 'EVTCHOP_RESET',
    11: 'EVTCHOP_INIT_CONTROL',
    12: 'EVTCHOP_EXPAND_ARRAY',
    13: 'EVTCHOP_SET_PRIORITY'
}

hvmop_commands = {
    0: 'HVMOP_SET_PARAM',
    1: 'HVMOP_GET_PARAM',
    9: 'HVMOP_PAGETABLE_DYING',
    15: 'HVMOP_GET_MEMTYPE'
}

xenpmuop_commands = {
    0: 'XENPMU_MODE_GET',
    1: 'XENPMU_MODE_SET',
    2: 'XENPMU_FEATURE_GET',
    3: 'XENPMU_FEATURE_SET',
    4: 'XENPMU_INIT',
    5: 'XENPMU_FINISH',
    6: 'XENPMU_LVTPC_SET',
    7: 'XENPMU_FLUSH'
}


def set_variables(varlist):
    import re
    for i in varlist:
        var = re.sub(r'([a-z])([A-Z]+)', r'\1_\2', varlist[i])
        var = var.upper()
        exec(f'global {var}; {var}={i:d}')


set_variables(hypercall_names)
set_variables(xenmem_commands)
set_variables(xen_version_commands)
set_variables(console_io_commands)
set_variables(schedop_commands)
set_variables(evtchop_commands)
set_variables(hvmop_commands)
set_variables(xenpmuop_commands)
set_variables({get_iverr(i): hypercall_status_codes[i][0] for i in hypercall_status_codes.keys()})



# CHIPSEC: Platform Security Assessment Framework
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#

import struct

from chipsec.helper.basehelper import Helper
from chipsec.library.exceptions import UnimplementedAPIError
from typing import Optional, TYPE_CHECKING
if TYPE_CHECKING:
    from ctypes import Array


class TestHelper(Helper):
    """Default test helper that emulates a Broadwell architecture.

    See datasheet for registers definition.
    http://www.intel.com/content/www/us/en/chipsets/9-series-chipset-pch-datasheet.html
    """

    def __init__(self):
        super(TestHelper, self).__init__()
        self.os_system = "test_helper"
        self.os_release = "0"
        self.os_version = "0"
        self.os_machine = "test"
        self.driver_loaded = True
        self.name = "TestHelper"

    def create(self):
        return True

    def delete(self):
        return True

    def start(self):
        return True

    def stop(self):
        return True

    def _generate_size_ffs(self, size: int) -> int:
        return ~(~0xFF << (size-1) * 8)

     # This will be used to probe the device, fake a Broadwell CPU
    def read_pci_reg(self, bus, device, function, address, size):
        if (bus, device, function) == (0, 0, 0):
            if size == 1:
                return 0x86
            elif size == 2:
                return 0x8086
            else:
                return 0x16008086
        elif (bus, device, function) == (0, 0x1f, 0):
            if size == 1:
                return 0x86
            elif size == 2:
                return 0x8086
            else:
                return 0x9D438086
        else:
            return self._generate_size_ffs(size)

    def get_threads_count(self):
        return 2

    def cpuid(self, eax, ecx):
        return 0x406F1, 0, 0, 0

    def write_pci_reg(self, bus, device, function, address, value, size):
        raise UnimplementedAPIError('write_pci_reg')

    def get_info(self):
        return self.name, self.driverpath

    def read_mmio_reg(self, phys_address, size):
        raise UnimplementedAPIError('read_mmio_reg')

    def write_mmio_reg(self, phys_address, size, value):
        raise UnimplementedAPIError('write_mmio_reg')

    def read_phys_mem(self, phys_address, length):
        raise UnimplementedAPIError('read_phys_mem')

    def write_phys_mem(self, phys_address, length, buf):
        raise UnimplementedAPIError('write_phys_mem')

    def alloc_phys_mem(self, length, max_phys_address):
        raise UnimplementedAPIError('alloc_phys_mem')

    def free_phys_mem(self, physical_address):
        raise UnimplementedAPIError('free_phys_mem')

    def va2pa(self, va):
        raise UnimplementedAPIError('va2pa')

    def map_io_space(self, physical_address, length, cache_type):
        raise UnimplementedAPIError('map_io_space')

    def read_io_port(self, io_port, size):
        raise UnimplementedAPIError('read_io_port')

    def write_io_port(self, io_port, value, size):
        raise UnimplementedAPIError('write_io_port')

    def read_cr(self, cpu_thread_id, cr_number):
        raise UnimplementedAPIError('read_cr')

    def write_cr(self, cpu_thread_id, cr_number, value):
        raise UnimplementedAPIError('write_cr')

    def read_msr(self, cpu_thread_id, msr_addr):
        raise UnimplementedAPIError('read_msr')

    def write_msr(self, cpu_thread_id, msr_addr, eax, edx):
        raise UnimplementedAPIError('write_msr')

    def load_ucode_update(self, cpu_thread_id, ucode_update_buf):
        raise UnimplementedAPIError('load_ucode_update')

    def get_descriptor_table(self, cpu_thread_id, desc_table_code):
        raise UnimplementedAPIError('get_descriptor_table')

    def EFI_supported(self):
        raise UnimplementedAPIError('EFI_supported')

    def get_EFI_variable(self, name, guid):
        raise UnimplementedAPIError('get_EFI_variable')

    def set_EFI_variable(self, name, guid, buffer, buffer_size, attrs):
        raise UnimplementedAPIError('set_EFI_variable')

    def delete_EFI_variable(self, name, guid):
        raise UnimplementedAPIError('delete_EFI_variable')

    def list_EFI_variables(self):
        raise UnimplementedAPIError('list_EFI_variables')

    def get_ACPI_table(self, table_name: str) -> Optional['Array']:
        raise UnimplementedAPIError('get_ACPI_table')

    def enum_ACPI_tables(self) -> Optional['Array']:
        raise UnimplementedAPIError('enum_ACPI_table')

    def msgbus_send_read_message(self, mcr, mcrx):
        raise UnimplementedAPIError('msgbus_send_read_message')

    def msgbus_send_write_message(self, mcr, mcrx, mdr):
        raise UnimplementedAPIError('msgbus_send_write_message')

    def msgbus_send_message(self, mcr, mcrx, mdr):
        raise UnimplementedAPIError('msgbus_send_message')

    def get_affinity(self):
        raise UnimplementedAPIError('get_affinity')

    def set_affinity(self, value):
        raise UnimplementedAPIError('set_affinity')

    def send_sw_smi(self, cpu_thread_id, SMI_code_data, _rax, _rbx, _rcx, _rdx, _rsi, _rdi):
        raise UnimplementedAPIError('send_sw_smi')

    def hypercall(self, rcx, rdx, r8, r9, r10, r11, rax, rbx, rdi, rsi, xmm_buffer):
        raise UnimplementedAPIError('hypercall')

    def getcwd(self):
        raise UnimplementedAPIError('getcwd')

    def retpoline_enabled(self) -> bool:
        return False


class ACPIHelper(TestHelper):
    """Generic ACPI emulation

    Emulates an RSDP that points to an RSDT by default.

    An XSDT is also provided.

    The provided RSDT and XSDT will contain the entries specified
    in RSDT_ENTRIES and XSDT_ENTRIES.

    RSDP will be placed inside EBDA which by default is located
    at 0x96000. In particular, RSDP can be found in 0x96100.

    Three regions are defined:
      * RSDP table [0x96100, 0x96114] or [0x96100, 0x96124]
      * XSDT table [0x100, 0x124 + 8 * len(XSDT_ENTRIES)]
      * RSDT table [0x200, 0x224 + 4 * len(RSDT_ENTRIES)]
    """
    USE_RSDP_REV_0 = True
    TABLE_HEADER_SIZE = 36

    EBDA_ADDRESS = 0x96000
    EBDA_PADDING = 0x100
    RSDP_ADDRESS = EBDA_ADDRESS + EBDA_PADDING

    XSDT_ADDRESS = 0x100
    RSDT_ADDRESS = 0x200

    def _create_rsdp(self):
        rsdp = b""
        if self.USE_RSDP_REV_0:
            # Emulate initial version of RSDP described in ACPI v1.0
            rsdp = (b"RSD PTR " +                            # Signature
                    struct.pack("<B", 0x1) +                # Checksum
                    b"TEST00" +                              # OEMID
                    struct.pack("<B", 0x0) +                # Revision
                    struct.pack("<I", self.RSDT_ADDRESS))   # RSDT Address
        else:
            # Emulate RSDP described in ACPI v2.0 onwards
            rsdp = (b"RSD PTR " +                            # Signature
                    struct.pack("<B", 0x1) +                # Checksum
                    b"TEST00" +                              # OEMID
                    struct.pack("<B", 0x2) +                # Revision
                    struct.pack("<I", self.RSDT_ADDRESS) +  # RSDT Address
                    struct.pack("<I", 0x24) +               # Length
                    struct.pack("<Q", self.XSDT_ADDRESS) +  # XSDT Address
                    struct.pack("<B", 0x0) +                # Extended Checksum
                    b"AAA")                                  # Reserved
        return rsdp

    def _create_generic_acpi_table_header(self, signature, length):
        return (signature +                  # Signature
                struct.pack("<I", length) +  # Length
                struct.pack("<B", 0x1) +     # Revision
                struct.pack("<B", 0x1) +     # Checksum
                b"OEMIDT" +                   # OEMID
                b"OEMTBLID" +                 # OEM Table ID
                b"OEMR" +                     # OEM Revision
                b"CRID" +                     # Creator ID
                b"CRRV")                      # Creator Revision

    def _create_rsdt(self):
        rsdt_length = self.TABLE_HEADER_SIZE + 4 * len(self.rsdt_entries)
        rsdt = self._create_generic_acpi_table_header(b"RSDT", rsdt_length)
        for rsdt_entry in self.rsdt_entries:
            rsdt += struct.pack("<I", rsdt_entry)
        return rsdt

    def _create_xsdt(self):
        xsdt_length = self.TABLE_HEADER_SIZE + 8 * len(self.xsdt_entries)
        xsdt = self._create_generic_acpi_table_header(b"XSDT", xsdt_length)
        for xsdt_entry in self.xsdt_entries:
            xsdt += struct.pack("<Q", xsdt_entry)
        return xsdt

    def _add_entry_to_rsdt(self, entry):
        self.rsdt_entries.append(entry)
        self.rsdt_descriptor = self._create_rsdt()

    def _add_entry_to_xsdt(self, entry):
        self.xsdt_entries.append(entry)
        self.xsdt_descriptor = self._create_xsdt()

    def __init__(self):
        super(ACPIHelper, self).__init__()
        self.rsdt_entries = []
        self.xsdt_entries = []
        self.rsdp_descriptor = self._create_rsdp()
        self.rsdt_descriptor = self._create_rsdt()
        self.xsdt_descriptor = self._create_xsdt()

    def read_phys_mem(self, pa, length):
        pa_lo = pa & 0xFFFFFFFF
        if pa_lo == 0x40E:
            return struct.pack("<H", self.EBDA_ADDRESS >> 4)
        elif (pa_lo >= self.EBDA_ADDRESS and
              pa_lo < self.RSDP_ADDRESS + len(self.rsdp_descriptor)):
            mem = b"\x00" * self.EBDA_PADDING + self.rsdp_descriptor
            offset = pa_lo - self.EBDA_ADDRESS
            return mem[offset:offset + length]
        elif pa_lo == self.RSDT_ADDRESS:
            return self.rsdt_descriptor[:length]
        elif pa_lo == self.XSDT_ADDRESS:
            return self.xsdt_descriptor[:length]
        else:
            return b"\xFF" * length


class DSDTParsingHelper(ACPIHelper):
    """Test helper containing generic descriptors for RSDP, RSDT and FADT

    Default entry for DSDT and X_DSDT inside FADT is 0x0

    One additional region is defined:
      * FADT table [0x400, 0x514]
    """
    USE_FADT_WITH_X_DSDT = True

    FADT_ADDRESS = 0x400
    DSDT_ADDRESS = 0x0
    X_DSDT_ADDRESS = 0x0

    def _add_fadt_to_sdt_entries(self):
        if self.USE_RSDP_REV_0:
            self._add_entry_to_rsdt(self.FADT_ADDRESS)
        else:
            self._add_entry_to_xsdt(self.FADT_ADDRESS)

    def _create_fadt(self):
        fadt = b""
        if self.USE_FADT_WITH_X_DSDT:
            fadt = self._create_generic_acpi_table_header(b"FACP", 0x10C)
            fadt += struct.pack("<I", 0x500)                # Address of FACS
            fadt += struct.pack("<I", self.DSDT_ADDRESS)    # DSDT
            fadt += struct.pack("<B", 0x1) * 96             # Padding
            fadt += struct.pack("<Q", self.X_DSDT_ADDRESS)  # X_DSDT
            fadt += struct.pack("<B", 0x1) * 120            # Remaining fields
        else:
            fadt = self._create_generic_acpi_table_header(b"FACP", 0x74)
            fadt += struct.pack("<I", 0x500)                # Address of FACS
            fadt += struct.pack("<I", self.DSDT_ADDRESS)    # DSDT
            fadt += struct.pack("<B", 0x1) * 72             # Remaining fields
        return fadt

    def __init__(self):
        super(DSDTParsingHelper, self).__init__()
        self._add_fadt_to_sdt_entries()
        self.fadt_descriptor = self._create_fadt()

    def read_phys_mem(self, pa, length):
        pa_lo = pa & 0xFFFFFFFF
        if pa_lo == self.FADT_ADDRESS:
            return self.fadt_descriptor[:length]
        else:
            parent = super(DSDTParsingHelper, self)
            return parent.read_phys_mem(pa, length)


class SPIHelper(TestHelper):
    """Generic SPI emulation

    Two regions are defined:
      * The flash descriptor [0x1000, 0x1FFF]
      * The BIOS image [0x2000, 0x2FFF]
    """
    RCBA_ADDR = 0xFED0000
    SPIBAR_ADDR = RCBA_ADDR + 0x3800
    SPIBAR_END = SPIBAR_ADDR + 0x200
    HSFS = SPIBAR_ADDR + 0x4
    FRAP = SPIBAR_ADDR + 0x50
    FREG0 = SPIBAR_ADDR + 0x54
    LPC_BRIDGE_DEVICE = (0, 0x1F, 0)

    def read_pci_reg(self, bus, device, function, address, size):
        if (bus, device, function) == self.LPC_BRIDGE_DEVICE:
            if address == 0xF0:
                return self.RCBA_ADDR
            elif address == 0xDC:
                return 0xDEADBEEF
            elif address == 0x0:
                return 0xAAAA8086
        return super(SPIHelper, self).read_pci_reg(bus, device,
                                                       function,
                                                       address, size)

    def read_mmio_reg(self, pa, size):
        if pa == self.FREG0:
            return 0x00010001
        elif pa == self.FREG0 + 4:
            return 0x00020002
        elif pa == self.FRAP:
            return 0xEEEEEEEE
        elif pa == self.HSFS:
            return (1 << 14)  # FDV = 1, the flash descriptor is valid
        elif pa >= self.SPIBAR_ADDR and pa < self.SPIBAR_END:
            return 0x0
        else:
            raise Exception("Unexpected address")

    def write_mmio_reg(self, pa, size, value):
        if pa < self.SPIBAR_ADDR or pa > self.SPIBAR_END:
            raise Exception("Write to outside of SPIBAR")

    def map_io_space(self, base, size, cache_type):
        raise UnimplementedAPIError("Not implemented")


class ValidChipsetHelper(TestHelper):
    def read_pci_reg(self, bus, device, function, address, size):
        if (bus, device, function) == (0, 0, 0):
            if size == 1:
                return 0x86
            elif size == 2:
                return 0x8086
            else:
                return 0x19048086
        elif (bus, device, function) == (0, 0x1f, 0):
            if size == 1:
                return 0x86
            elif size == 2:
                return 0x8086
            else:
                return 0x9D438086
        else:
            return self._generate_size_ffs(size)


class InvalidChipsetHelper(TestHelper):
    def read_pci_reg(self, bus, device, function, address, size):
        if (bus, device, function) == (0, 0, 0):
            if size == 1:
                return 0x86
            elif size == 2:
                return 0x8086
            else:
                return 0xBEEF8086
        elif (bus, device, function) == (0, 0x1f, 0):
            if size == 1:
                return 0x86
            elif size == 2:
                return 0x8086
            else:
                return 0x9D438086
        else:
            return self._generate_size_ffs(size)

    def cpuid(self, eax, ecx):
        return 0xfffff, 0, 0, 0


class InvalidPchHelper(TestHelper):
    def read_pci_reg(self, bus, device, function, address, size):
        if (bus, device, function) == (0, 0, 0):
            if size == 1:
                return 0x86
            elif size == 2:
                return 0x8086
            else:
                return 0x19048086
        elif (bus, device, function) == (0, 0x1f, 0):
            if size == 1:
                return 0x86
            elif size == 2:
                return 0x8086
            else:
                return 0xBEEF8086
        else:
            return self._generate_size_ffs(size)



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
Access to CPU resources (for each CPU thread): Model Specific Registers (MSR), IDT/GDT

usage:
    >>> read_msr( 0x8B )
    >>> write_msr( 0x79, 0x12345678 )
    >>> get_IDTR( 0 )
    >>> get_GDTR( 0 )
    >>> dump_Descriptor_Table( 0, DESCRIPTOR_TABLE_CODE_IDTR )
    >>> IDT( 0 )
    >>> GDT( 0 )
    >>> IDT_all()
    >>> GDT_all()
"""

from typing import Dict, Tuple, Optional
from chipsec.library.logger import logger, print_buffer_bytes


DESCRIPTOR_TABLE_CODE_IDTR = 0
DESCRIPTOR_TABLE_CODE_GDTR = 1
DESCRIPTOR_TABLE_CODE_LDTR = 2

MTRR_MEMTYPE_UC = 0x0
MTRR_MEMTYPE_WC = 0x1
MTRR_MEMTYPE_WT = 0x4
MTRR_MEMTYPE_WP = 0x5
MTRR_MEMTYPE_WB = 0x6
MemType: Dict[int, str] = {
    MTRR_MEMTYPE_UC: 'Uncacheable (UC)',
    MTRR_MEMTYPE_WC: 'Write Combining (WC)',
    MTRR_MEMTYPE_WT: 'Write-through (WT)',
    MTRR_MEMTYPE_WP: 'Write-protected (WP)',
    MTRR_MEMTYPE_WB: 'Writeback (WB)'
}


class Msr:

    def __init__(self, cs):
        self.helper = cs.helper
        self.cs = cs

    def get_cpu_thread_count(self) -> int:
        thread_count = self.helper.get_threads_count()
        if thread_count is None or thread_count < 0:
            logger().log_hal("helper.get_threads_count didn't return anything. Reading MSR 0x35 to find out number of logical CPUs (use CPUID Leaf B instead?)")
            thread_count = self.cs.register.read_field("IA32_MSR_CORE_THREAD_COUNT", "Thread_Count")

        if 0 == thread_count:
            thread_count = 1
        logger().log_hal(f'[cpu] # of logical CPUs: {thread_count:d}')
        return thread_count

    # @TODO: fix
    def get_cpu_core_count(self) -> int:
        core_count = self.cs.register.read_field("IA32_MSR_CORE_THREAD_COUNT", "Core_Count")
        return core_count

##########################################################################################################
#
# Read/Write CPU MSRs
#
##########################################################################################################

    def read_msr(self, cpu_thread_id: int, msr_addr: int) -> Tuple[int, int]:
        (eax, edx) = self.helper.read_msr(cpu_thread_id, msr_addr)
        logger().log_hal(f'[cpu{cpu_thread_id:d}] RDMSR( 0x{msr_addr:x} ): EAX = 0x{eax:08X}, EDX = 0x{edx:08X}')
        return (eax, edx)

    def write_msr(self, cpu_thread_id: int, msr_addr: int, eax: int, edx: int) -> None:
        self.helper.write_msr(cpu_thread_id, msr_addr, eax, edx)
        logger().log_hal(f'[cpu{cpu_thread_id:d}] WRMSR( 0x{msr_addr:x} ): EAX = 0x{eax:08X}, EDX = 0x{edx:08X}')
        return None

##########################################################################################################
#
# Get CPU Descriptor Table Registers (IDTR, GDTR, LDTR..)
#
##########################################################################################################

    def get_Desc_Table_Register(self, cpu_thread_id: int, code: int) -> Tuple[int, int, int]:
        desc_table = self.helper.get_descriptor_table(cpu_thread_id, code)
        if desc_table is None:
            logger().log_hal(f'[msr] Unable to locate CPU Descriptor Table: Descriptor table code = {code:d}')
            return (0, 0, 0)
        return desc_table

    def get_IDTR(self, cpu_thread_id: int) -> Tuple[int, int, int]:
        (limit, base, pa) = self.get_Desc_Table_Register(cpu_thread_id, DESCRIPTOR_TABLE_CODE_IDTR)
        logger().log_hal(f'[cpu{cpu_thread_id:d}] IDTR Limit = 0x{limit:04X}, Base = 0x{base:016X}, Physical Address = 0x{pa:016X}')
        return (limit, base, pa)

    def get_GDTR(self, cpu_thread_id: int) -> Tuple[int, int, int]:
        (limit, base, pa) = self.get_Desc_Table_Register(cpu_thread_id, DESCRIPTOR_TABLE_CODE_GDTR)
        logger().log_hal(f'[cpu{cpu_thread_id:d}] GDTR Limit = 0x{limit:04X}, Base = 0x{base:016X}, Physical Address = 0x{pa:016X}')
        return (limit, base, pa)

    def get_LDTR(self, cpu_thread_id: int) -> Tuple[int, int, int]:
        (limit, base, pa) = self.get_Desc_Table_Register(cpu_thread_id, DESCRIPTOR_TABLE_CODE_LDTR)
        logger().log_hal(f'[cpu{cpu_thread_id:d}] LDTR Limit = 0x{limit:04X}, Base = 0x{base:016X}, Physical Address = 0x{pa:016X}')
        return (limit, base, pa)

##########################################################################################################
#
# Dump CPU Descriptor Tables (IDT, GDT, LDT..)
#
##########################################################################################################

    def dump_Descriptor_Table(self, cpu_thread_id: int, code: int, num_entries: Optional[int] = None) -> Tuple[int, int]:
        (limit, _, pa) = self.helper.get_descriptor_table(cpu_thread_id, code)
        dt = self.helper.read_phys_mem(pa, limit + 1)
        total_num = len(dt) // 16
        if (num_entries is None) or (total_num < num_entries):
            num_entries = total_num
        logger().log(f'[cpu{cpu_thread_id:d}] Physical Address: 0x{pa:016X}')
        logger().log(f'[cpu{cpu_thread_id:d}] # of entries    : {total_num:d}')
        logger().log(f'[cpu{cpu_thread_id:d}] Contents ({num_entries:d} entries):')
        print_buffer_bytes(dt)
        logger().log('--------------------------------------')
        logger().log('#    segment:offset         attributes')
        logger().log('--------------------------------------')
        for i in range(0, num_entries):
            offset = (dt[i * 16 + 11] << 56) | (dt[i * 16 + 10] << 48) | (dt[i * 16 + 9] << 40) | (dt[i * 16 + 8] << 32) | (dt[i * 16 + 7] << 24) | (dt[i * 16 + 6] << 16) | (dt[i * 16 + 1] << 8) | dt[i * 16 + 0]
            segsel = (dt[i * 16 + 3] << 8) | dt[i * 16 + 2]
            attr = (dt[i * 16 + 5] << 8) | dt[i * 16 + 4]
            logger().log(f'{i:03d}  {segsel:04X}:{offset:016X}  0x{attr:04X}')

        return (pa, dt)

    def IDT(self, cpu_thread_id: int, num_entries: Optional[int] = None) -> Tuple[int, int]:
        logger().log_hal(f'[cpu{cpu_thread_id:d}] IDT:')
        return self.dump_Descriptor_Table(cpu_thread_id, DESCRIPTOR_TABLE_CODE_IDTR, num_entries)

    def GDT(self, cpu_thread_id: int, num_entries: Optional[int] = None) -> Tuple[int, int]:
        logger().log_hal(f'[cpu{cpu_thread_id:d}] GDT:')
        return self.dump_Descriptor_Table(cpu_thread_id, DESCRIPTOR_TABLE_CODE_GDTR, num_entries)

    def IDT_all(self, num_entries: Optional[int] = None) -> None:
        for tid in range(self.get_cpu_thread_count()):
            self.IDT(tid, num_entries)

    def GDT_all(self, num_entries: Optional[int] = None) -> None:
        for tid in range(self.get_cpu_thread_count()):
            self.GDT(tid, num_entries)
